/*
 * TATransition.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 2-ott-03
 */
package xrel.analyzer;

import java.util.HashSet;
import java.util.Iterator;

import xrel.parser.SymElement;

/**
 * Defines a normal tree-transition from a state to another.
 * @author Fabrizio Bisi
 */
class TATransition {
	TAState src; // from state Src
	TAState dst; // to state Dst
	String content; // automaton name that matches with the content of the tag
	HashSet variables; // variables to bind in transition
	String label; // the label 
	HashSet diffLabels;
	
	/**
	 * Initializes a (labeled) tree-transition from a state to another.
	 * @param lbl the label of the tag
	 * @param s the source state
	 * @param d the destination state
	 * @param c the name of the pattern that represents the content of the tag
	 * @param var the set of variables to bind in the matching
	 */
	public TATransition(String lbl, TAState s, TAState d, String c,
					  HashSet var) {
		label = lbl;
		src = s;
		dst = d;
		content = c;
		variables = var;
		diffLabels = new HashSet();
	}

	/**
	 * Gives the name of the transition.
	 * @return the string that represents the transition
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return src + " = " + getLabel() + "[" + content + "]  --> " +
			dst + "  " + getVariablesString(); 
	}
	
	private String getLabel() {
		if (diffLabels.size() == 0)
			return label;
		else 
			return "(" + label + "\\" + diffLabels + ")"; 
	}
	
	private String getVariablesString() {
		String rslt = "{";
		if (variables != null) {
			Iterator itr = variables.iterator();
			boolean first = true;
			while (itr.hasNext()) {
				SymElement el = (SymElement)itr.next();
				if (first) first = false;
				else rslt = rslt + ",";
				rslt = rslt + el.getUniqueName();
			}
		}
		return rslt + "}";
	}
	
	public void normalize(HashSet labels) {
		diffLabels.clear();
		diffLabels.addAll(labels);
	}
}
