/* SimpleNode.java - modified */

package xrel.parser;

import java.io.PrintStream;
import java.util.HashSet;
import java.util.Iterator;

import xrel.XreError;
import xrel.xml.Builder;
//import java.lang.reflect.Constructor;

public class SimpleNode implements Node,Cloneable {
  protected Node parent;
  protected Node[] children;
  protected int id;
  protected XRelParser parser;
	
	/**
	 * The standard constructor. It accepts one only argument: the constant
	 * that identifies the subclass.
	 * @param i the constant that identifies the subclass
	 */
  public SimpleNode(int i) {
//  	XSet = new HashSet();
  	Vars = new HashSet();
//  	VarMap = null;
    id = i;
  }
	
	/**
	 * This constructor lets you to specify the parser.
	 * @param p the parser
	 * @param i the constant that identifies the subclass
	 */
  public SimpleNode(XRelParser p, int i) {
    this(i);
    parser = p;
  }

  public void jjtSetParent(Node n) { parent = n; }
  public Node jjtGetParent() { return parent; }
	
	/**
	 * Adds n to the children list of the current node. Specifically if "i" 
	 * is inside the range of the existing children it will replace the i-th 
	 * node with n, while if "i" is greater than the index of the last child 
	 * it will append the specified node n to the existing children to the
	 * i-th position. Note that the children list will be expanded to contain
	 * however at least i+1 nodes. 
	 * @param n the node to become child of the current node
	 * @param i the position in the children list where to put n 
	 * @see xrel.parser.Node#jjtAddChild(xrel.parser.Node, int)
	 */
  public void jjtAddChild(Node n, int i) {
    if (children == null) {
      children = new Node[i + 1];
    } 
    else if (i >= children.length) {
      Node c[] = new Node[i + 1];
      System.arraycopy(children, 0, c, 0, children.length);
      children = c;
    }
//    else {
//    	/* Checks if the replaced node was the root of some variables
//    	 * and in this case change these variables
//    	 * TODO try to replace this code
//    	 */
//    	SimpleNode oldNode = (SimpleNode)children[i]; 
//    	if ((oldNode != null) && (oldNode.XSet.size() > 0)) {
//    		Iterator itr = oldNode.XSet.iterator();
//    		while (itr.hasNext()) {
//    			SymElement el = (SymElement)itr.next();
//    			// TODO the use of symtab is temporary!
//					System.out.println("\n>>>>>>>>>>>>>>>>>>>>>>>>>> should change " + ((ASTAs)this).getIdentifier());
////    			oldNode.symtab.changeVarNode(el,(SimpleNode)n);
//    		}
//    	}
//    }
    children[i] = n;
  }

  public Node jjtGetChild(int i) {
    return children[i];
  }

  public int jjtGetNumChildren() {
    return (children == null) ? 0 : children.length;
  }

  /** Accept the visitor. */
  public Object jjtAccept(XRelParserVisitor visitor, Object data) {
    return visitor.visit(this, data);
  }

  /** Accept the visitor. */
  public Object childrenAccept(XRelParserVisitor visitor, Object data) {
    if (children != null) {
      for (int i = 0; i < children.length; ++i) {
        children[i].jjtAccept(visitor, data);
      }
    }
    return data;
  }
	
	/**
	 * Returns a string representation of the node.
	 * You can override this method in subclasses of SimpleNode to
	 * customize the way the node appears when the tree is dumped.  If
	 * your output uses more than one line you should override
	 * toString(String), otherwise overriding toString() is probably all
	 * you need to do.
	 * @return the string associated to the node
	 * @see java.lang.Object#toString()
	 */
  public String toString() { 
  	return XRelParserTreeConstants.jjtNodeName[id]; 
  }
  /**
	 * Returns a string representation of the node.
   * You can override this method in subclasses of SimpleNode to
   * customize the way the node appears when the tree is dumped.  It's
   * specially useful when your output uses more than one line (if
   * your output uses one only line overriding toString() is probably all
   * you need to do).
   * @param prefix a prefix for the string
   * @return the string associated to the node
   */
  public String toString(String prefix) { 
  	return prefix + toString(); 
  }

  /************************* Modified *******************/

  protected Token first, last; // new
	
	// overriden
  public void jjtOpen() {
		first = parser.getToken(1);	
  }
	
	// overriden
  public void jjtClose() {
		last = parser.getToken(0);	
  }
	
	/** Returns the first token associated to this node. Note that some
	 * transformation on the the syntax tree loose this information.
	 * @return the first token
	 */
  public Token getFirstToken() { return first; } // new
  /** Returns the last token associated to this node. Note that
   * some transformations on the syntax tree loose this information.
   * @return the last token
   */
  public Token getLastToken() { return last; }   // new
	
  /************************* Added *******************/
	
//	/** the set of variables that have this node as root */
//	protected HashSet XSet;
//	/** a reference to the symbol table. Used by automata & output values */
//	protected SymTable symtab; 
	/** variables annotated on the value. Used by automata */
	protected HashSet Vars;
	
  /**
   * <p>Writes out to the stream "out" the subtree of this node
   * in a human-readable structured format. 
   * While dumpIntTree() is used to write the original syntax tree,
   * this function is used to write XML values or output expressions. </p> 
	 * <p>N.B.: 
	 * <ul>
	 * <li>it can only show values/expressions </li>
	 * <li>it correctly handles only values/expressions without ground types</li>
	 * </ul></p>
   * @see #dumpIntTree(PrintStream out, String prefix)
   * @param out the stream where to dump out
   * @param prefix a string of prefix 
   */
  public void writeExpression(PrintStream out, String prefix, SymTable st) {
		if (getClass() == ASTTag.class) {
			String label = ((ASTTag)this).getName();
			if (label.startsWith(SpecialNames.StrLiteralPrefix) && 
				label.endsWith(SpecialNames.StrLiteralSuffix)) {
			  // WARNING! Code AD HOC!
				out.print(label.substring(1,label.length()));
			}
			else {
				SimpleNode n = (SimpleNode)jjtGetChild(0);
				boolean isLeaf = n.isLeaf();
				out.print(label + "[");
				if (!isLeaf) out.print(" ");
				n.writeExpression(out,prefix + " ",st);
				if (!isLeaf) out.print(" ");
				out.print("]");
			}
		}
		else if (getClass() == ASTComma.class) {
			if (parent != null && parent.getClass() != ASTTag.class)
				out.print("( ");
			boolean first = true;
			for (int i= 0; i< jjtGetNumChildren(); i++) {
				if (first) first = false;
				else out.print(", ");
				((SimpleNode)jjtGetChild(i)).writeExpression(out,prefix,st);
			}
			if (parent != null && parent.getClass() != ASTTag.class)
				out.print(" )");
		}
	  else if (getClass() == ASTEmpty.class) {
			if (parent == null || parent.getClass() != ASTTag.class)
				out.print("()");
	  }
	  else if (getClass() == ASTStringLiteral.class) {
			String label = ((ASTStringLiteral)this).getValue();
			out.print(label);
	  }
		else if (getClass() == ASTIdentifier.class) {
			String id = ((ASTIdentifier)this).getName();
			int scope = ((ASTIdentifier)this).getScope();
			SymElement el = st.search(SymElementKind.VARIABLE,scope,id);
			if (el == null)
				el = st.search(SymElementKind.VARIABLE,0,id);
			if (el != null) {
				SimpleNode n = el.getValue();
				if (n != null) n.writeExpression(out,prefix,st);
				// else correctly handle null values doing nothing!
				//else
				//	XreError.showbug_and_exit("null value?");
			}
			else XreError.showerror_and_exit("can't find identifier " + id);
//			else {
//				el = st.search(SymElementKind.S_EXPRESSIONS,id);
//				if (el != null) {
//					SimpleNode n = el.getNode();
//					if (n != null) n.writeExpression(out,prefix,st);
//				}
//				else XreError.showerror_and_exit("can't find identifier " + id);
//			}
		}
	  else {
			XreError.showbug_and_exit("it works only with expressions, sorry");
	  }
	}

	
	/**
	 * <p>Exports to the stream "out" the subtree of this node
	 * in an XML format. </p> 
   * <p>N.B.: 
   * <ul>
   * <li>it can only show values/expressions </li>
   * <li>it correctly handles only values/expressions without ground types</li>
   * </ul></p>
	 * @param out the stream where to dump out
	 * @param prefix a string of prefix 
	 */
	public void exportExpression(String filename, SymTable st) {
		xrel.xml.Builder wr = new xrel.xml.Builder();
		wr.openDoc(filename,"Data");
		exportXMLFragment(wr,st);
		wr.closeDoc();
	}
	
	private void exportXMLFragment(Builder wr, SymTable st) {
	  if (getClass() == ASTTag.class) {
			String label = ((ASTTag)this).getName();
		  if (label.startsWith(SpecialNames.StrLiteralPrefix) && 
			    label.endsWith(SpecialNames.StrLiteralSuffix)) {
				// WARNING! Code AD HOC!
			  wr.writeText(label.substring(1,label.length()));
		  }
		  else {
				SimpleNode n = (SimpleNode)jjtGetChild(0);
				//boolean isLeaf = n.isLeaf();
				wr.openNode(label);
			  n.exportXMLFragment(wr,st);
			  wr.closeNode(label);
		  }
	  }
	  else if (getClass() == ASTComma.class) {
		  for (int i= 0; i< jjtGetNumChildren(); i++) {
				SimpleNode n = (SimpleNode)jjtGetChild(i);
				n.exportXMLFragment(wr,st);
		  }
	  }
		else if (getClass() == ASTEmpty.class) {
			// do nothing
		}
		else if (getClass() == ASTStringLiteral.class) {
			String label = ((ASTStringLiteral)this).getValue();
			wr.writeText(label);
		}
		else if (getClass() == ASTIdentifier.class) {
			String id = ((ASTIdentifier)this).getName();
			int scope = ((ASTIdentifier)this).getScope();
			SymElement el = st.search(SymElementKind.VARIABLE,scope,id);
			if (el == null) el = st.search(SymElementKind.VARIABLE,0,id);
			if (el != null) {
				SimpleNode n = el.getValue();
				if (n != null) n.exportXMLFragment(wr,st);
				// do nothing
				//else XreError.showbug_and_exit("null value?");
			}
			else XreError.showerror_and_exit("can't find variable " + id);
		}
		else {
			XreError.showbug_and_exit("it works only with expressions, sorry");
		}
  }
	
	private boolean isLeaf() {
		String label = getClass() == ASTTag.class ? ((ASTTag)this).getName() : "";
		return
			(getClass() == ASTEmpty.class) ||
			(getClass() == ASTTag.class && 
				((label.startsWith(SpecialNames.StrLiteralPrefix) && 
					label.endsWith(SpecialNames.StrLiteralSuffix)) || 
					label.equals(SpecialNames.StringTypeName))
				) || 
			(getClass() == ASTIdentifier.class) || 
			(getClass() == ASTStringLiteral.class) || 
			(getClass() == ASTString.class); 
	}
	
	/** 
	 * <p>Writes a tree in a compact human-readable form. </p>
	 * <p>N.B.: 
	 * <ul>
	 * <li>it can only show patterns, types, expressions (without variable
	 * extension) or values; it can't show show the whole original syntax 
	 * tree </li>
	 * <li>it can correctly handle both original patterns
	 * with ground types and modified patterns with ground types removed</li>
	 * </ul></p>
	 * @param out the stream where to print out
	 */
	public void writePattern(PrintStream out) {
		out.print(getPattern());
	}
	
	public String getPattern() {
		String result = "";
		if (getClass() == ASTComma.class) {
			if (parent != null && parent.getClass() != ASTTag.class) result += "( ";
			boolean first = true;
			for (int i= 0; i< jjtGetNumChildren(); i++) {
				if (first) first = false;
				else result += ", ";
				result += ((SimpleNode)jjtGetChild(i)).getPattern();
			}
			if (parent != null && parent.getClass() != ASTTag.class) result += " )";
		}
		else if (getClass() == ASTPar.class) {
			if (parent != null && parent.getClass() != ASTTag.class) result += "( ";
			boolean first = true;
			for (int i= 0; i< jjtGetNumChildren(); i++) {
				if (first) first = false;
				else result += " | ";
				result += ((SimpleNode)jjtGetChild(i)).getPattern();
			}
			if (parent != null && parent.getClass() != ASTTag.class) result += " )";
		}
		else if (getClass() == ASTTag.class) {
			String label = ((ASTTag)this).getName();
			if (label.startsWith(SpecialNames.StrLiteralPrefix) && 
			    label.endsWith(SpecialNames.StrLiteralSuffix)) {
			  // WARNING! Code AD HOC!
			  result += label.substring(1,label.length());
			}
			else if (label.equals(SpecialNames.StringTypeName)) {
				result += SpecialNames.StringName;
			}
			else {
				SimpleNode n = (SimpleNode)jjtGetChild(0);
				String nLabel = n.getClass() == ASTTag.class ? ((ASTTag)n).getName() : "";
				boolean isLeaf =
				  (n.getClass() == ASTEmpty.class) ||
					(n.getClass() == ASTTag.class && 
					  ((nLabel.startsWith(SpecialNames.StrLiteralPrefix) && 
							nLabel.endsWith(SpecialNames.StrLiteralSuffix)) || 
							nLabel.equals(SpecialNames.StringTypeName))
						) || 
					(n.getClass() == ASTIdentifier.class) || 
					(n.getClass() == ASTStringLiteral.class) || 
				  (n.getClass() == ASTString.class); 
				result += label + "[";
				if (!isLeaf) result += " ";
				result += n.getPattern();
				if (!isLeaf) result += " ";
				result += "]";
			}
		}
		else if (getClass() == ASTEmpty.class) {
			if (parent == null || parent.getClass() != ASTTag.class)
			result += "()";
		}
		else if (getClass() == ASTIdentifier.class) {
			String label = ((ASTIdentifier)this).getName();
			result += label;
		}
		else if (getClass() == ASTOperator.class) {
			//out.print("(");
			result += ((SimpleNode)jjtGetChild(0)).getPattern();
			//out.print(")");
			result += ((ASTOperator)this).getOperator();
		}
		else if (getClass() == ASTString.class) {
			result += SpecialNames.StringName;
		}
		else if (getClass() == ASTStringLiteral.class) {
			String label = ((ASTStringLiteral)this).getValue();
			result += label;
		}
		else if (getClass() == ASTAs.class) {
			//out.print("(");
			result += ((SimpleNode)jjtGetChild(0)).getPattern();
			//out.print(")");
			result += " as ";
			String id = ((ASTAs)this).getIdentifier();
			result += id;
		}
		else {
			XreError.showbug_and_exit("unknown or unsupported node type");
		}
		return result;
	}
	
	/**
	 * Writes out to the stream "out" the subtree of this node. 
	 * @param out the stream where to dump out
	 * @param prefix a string of prefix 
	 */
  public void dumpIntTree(PrintStream out, String prefix) {
		// Write this node
		out.println(toString(prefix));
		// Recursively write its children
		if (children != null) {
		  for (int i = 0; i < children.length; ++i) {
				SimpleNode n = (SimpleNode)children[i];
				if (n != null) { n.dumpIntTree(out,prefix + " "); }
		  }
		}
  }

	/** 
	 * <p>Writes a tree in a internal flat representation. 
	 * Flat means that description stands in a single row.
	 * Internal means that the internal tree structure is shown
	 * istead of a more human-readable format. </p>
	 * <p>For example "a[],b[] | c[]" is represented as
	 * <blockquote> 
	 * Par[ Comma[ Tag "a"[Empty], Tag "b"[Empty] ], Tag "c"[Empty] ]
	 * </blockquote></p>
	 * <p>N.B.: 
	 * <ul>
	 * <li>also if this function is used only to show
	 * patterns, types, expressions or values it could as well
	 * show the whole original syntax tree </li>
	 * </ul></p>
	 * @param out the stream where to print out
	 */
	public void dumpIntFlat(PrintStream out) {
		// Write this node
		out.print(this);
		// If there are children, write [ ch1, ... , chn ]
		if (jjtGetNumChildren() > 0) {
			// Open bracket
			out.print("[ ");	
			// Recursively write children separated by commas
			boolean isFirst = true;
			for (int i = 0; i < jjtGetNumChildren(); i++) {
				SimpleNode n = (SimpleNode)jjtGetChild(i);
				if (isFirst) isFirst = false;
				else out.print(", ");
				if (n != null) n.dumpIntFlat(out);
				else	XreError.showbug_and_exit("child null pointer");
			}
			// Close bracket
			out.print(" ]");
		}
	}

	/**
	 * Associate the node with the specified set of variables.
	 * A check is made if the node is a tag
	 * @param newSet
	 */
	public void annotateValue(HashSet newSet) {
		if (getClass() == ASTTag.class) Vars.addAll(newSet);
		else XreError.showbug_and_exit("attempting to annotate a node " +
		                               "that is not a tag");
	}
	
	/**
	 * <p>Cleans the subtree of the current node from variable annotations. 
	 * Note that it doesn't clean the annotations on the current node, 
	 * only on the nodes of the subtree.</p>
	 * @see #eraseAnnotations()
	 */
	public void eraseSubtreeAnnotations() {
		// Called by TASimple.matchSequence
		// There are problems if I clean the variables of the node itself 
		if (children != null) 
			for (int i = 0; i < children.length; i++) {
				SimpleNode child = (SimpleNode)children[i];
				if (child != null) child.eraseAnnotations();
			}
	}
	/**
	 * <p>Cleans the current node and its subtree from variable annotations. 
	 * Note that it cleans the annotations both from the current node and 
	 * from the nodes in the subtree.</p>
	 * @see #eraseSubtreeAnnotations()
	 */
	public void eraseAnnotations() {
		// Called by PatternMatching.matchWith
		// There are problems if I only clean the variables of the subtree 
		if (getClass() == ASTTag.class) Vars.clear();
		if (children != null) 
			for (int i = 0; i < children.length; i++) {
				SimpleNode child = (SimpleNode)children[i];
				if (child != null) child.eraseAnnotations();
			}
	}
	
	private String showvars() {
		String result = "";
		if (Vars.size() > 0) {
			Iterator it = Vars.iterator();
			result = " { ";
			boolean first = true;
			while(it.hasNext()) {
				SymElement el = (SymElement)it.next();
				String v = el.getUniqueName();
				if (first) first = false;
				else result += ", ";
				result += v; 
			}
			result += " }";
		}
		return result;
	}
	
	public void showAnnotations(PrintStream out, String prefix) {
	  out.println(toString(prefix) + showvars());
	  if (children != null) {
			for (int i = 0; i < children.length; ++i) {
			  SimpleNode n = (SimpleNode)children[i];
			  if (n != null) n.showAnnotations(out,prefix + " ");
		  }
		}
	}
	
	/* Calls bindVariables() on this node and on all the node in the
	 * subtree
	 */
	public void bindAll() {
		bindVariables();
		if (children != null) {
		  for (int i = 0; i < children.length; ++i) {
				SimpleNode n = (SimpleNode)children[i];
				if (n != null) n.bindAll();
			}
	  }
	}
	
	/* Associate the subtree of this node to the variables annotated 
	 * on the node itself. 
	 * Values are made of tags, sequences and leaves (empty nodes).
	 * Note that you can distinguish between unmatched variables and
	 * variables matched with the empty string as the former have a null
	 * value, while the latter have a value with an empty node.
	 * N.B.: note that the annotation algorithm don't annotate in any
	 * case an empty node, so that it's impossible that this function
	 * be called on an empty node, so that unmatched variables and variables
	 * matched with the empty string as actually indistinguishable
	 */
	private void bindVariables() {
		Iterator itr = Vars.iterator();
		// do a loop on all variables to bind
		while (itr.hasNext()) {
			// idEl is the element relative to the current variable to bind
			SymElement idEl = (SymElement)itr.next();
			// now see the old value associated to the variable
			SimpleNode oldVal = idEl.getValue();
			if (oldVal == null || oldVal.getClass() == ASTEmpty.class) {
				/* if no value is associated to the variable or if the old
				 * value associated to the variable is empty assign this
				 * piece of the input value to the variable.
				 * N.B.: there is no need to copy the value as we don't perform
				 * destructive operations on it, we only read it in the
				 * pattern matching and then we read it to compute the expressions
				 * in the code
				 */ 
				idEl.setValue(this);
			}
			else if (oldVal.getClass() == ASTComma.class) {
				// if the value is empty we don't anything
				if (this.getClass() != ASTEmpty.class) {
					// append the value as the last element of the sequence
					oldVal.jjtmAddChildExt(this,oldVal.jjtGetNumChildren());
				}
			}
			else if (oldVal.getClass() == ASTTag.class) {
				// if the value is empty we don't anything
				if (this.getClass() != ASTEmpty.class) {
					// if the old value is a tag we introduce a comma
					ASTComma newcomma = new ASTComma(XRelParserTreeConstants.JJTCOMMA);
					newcomma.jjtmAddChildExt(oldVal,0);
					newcomma.jjtmAddChildExt(this,1);
					idEl.setValue(newcomma);
				}
			}
			else {
				/* Hey! What is it? */
				XreError.showbug_and_exit("Internal bug");
			}
		}
	}

//	/**
//	 * Specifies that this node is the root node for an expression
//	 * that matches with the variable var.
//	 * @param var the variable that matches with this subtree
//	 */
//	public void addVariable(SymElement var) {
//		XSet.add(var);
//	}
	
	/**
	 * Searches a token with name tk from the token start.
	 * <P> N.B.: I've added this to search "as". </P>
	 * @param tk the name of the token to search
	 * @param start where to start to search
	 * @return the found token or null
	 */
  public Token searchTokenFrom(String tk, Token start) {
	Token result = null;
	Token c = start;
	do {
		if (c.image.equals(tk)) { 
			result = c; 
			break;
		} 
		c = c.next;
	}
	while (c != null);
	return result;
  }
	
  /* jjtm functions 
   * --------------
   * jjt Modified functions
   */
	
	/**
	 * <p>Clones this node and the whole subtree (by converse
   * clone() makes only a copy of the node itself). </p> 
	 * @return the root of the new subtree (the copy of the current node)
	 */
	public SimpleNode jjtmCloneSubtree() {
		// note that "this" can't be null
		SimpleNode thisCopy = (SimpleNode)clone();
		if (thisCopy != null) {
			if (this.children != null) {
				thisCopy.children = new Node[children.length];
				for (int i = 0; i < children.length; i ++) {
					SimpleNode child = (SimpleNode)children[i];
					SimpleNode childCopy = null;
					if (child != null) {
						childCopy = child.jjtmCloneSubtree();
					}
					else {
						XreError.showbug_and_exit("null child");
						return null;
					}
					//thisCopy.jjtmAddChildExt(childCopy,i);
					thisCopy.children[i] = childCopy;
					childCopy.jjtSetParent(thisCopy);
				}
			}
			return thisCopy;
		}
		else {
			// out of memory?
			XreError.showerror_and_exit("cannot duplicate tree (out of memory?)");
			return null;
		}
	}
	
	/**
	 * Clones this node. References to the parent and to the children are not
	 * changed.
	 * @return the copy of the object
	 * @see java.lang.Object#clone()
	 */
	protected Object clone() {
		Object o;
		try {
			o = super.clone();
//			((SimpleNode)o).XSet = (HashSet)XSet.clone();
			((SimpleNode)o).Vars = (HashSet)Vars.clone(); // it doesn't matter but..
//			if (VarMap != null)
//				((SimpleNode)o).VarMap = (HashMap)VarMap.clone();
		}
		catch(CloneNotSupportedException e) {
			XreError.showbug_and_exit("absurd! Object.clone() doesn't exist?");
			return null;
		}
		return o;
	}
	
	/**
	 * Calls jjtAddChild and sets the parent pointer of the child.
	 * You should use this function instead of jjtAddChild. 
	 * @param n the node that will be child of the current node
	 * @param i the position inside the children list.
	 */
	public void jjtmAddChildExt(Node n, int i) {
		jjtAddChild(n,i);
		n.jjtSetParent(this);
	}
	
	/**
	 * Moves children subset [from,upto] from this node to dst.
	 * @param dst node to which move
	 * @param from index of the first child to move
	 * @param upto index of the last child to move
	 */
  public void jjtmMoveChildren(Node dst, int from, int upto) {
		if (children != null && from < children.length) {
			if (upto > children.length-1) upto = children.length-1;
			for (int i = from; i <= upto; i++) {
				// append to the end of the children list of dst 
				((SimpleNode)dst).jjtmAddChildExt(children[i],dst.jjtGetNumChildren());
			}
			// remove children from source list
			int numToMove = upto - from +1;
			Node c[] = new Node[children.length - numToMove];
			if (from > 0) System.arraycopy(children, 0, c, 0, from);
			if (upto < children.length-1) 
				System.arraycopy(children, upto+1, c, from, children.length-upto-1);
			children = c;
		}
  }
  
  /**
   * Sets the first and the last token of the current node.
   * @param frst the first token 
   * @param lst the last token
   */
  public void jjtmSetTokens(Token frst, Token lst) {
  	first = frst;
  	last = lst;
  }
	
	/**
	 * Searches a child. When not found it returns -1.
	 * @param node the node to find
	 * @return the index of the child or -1 when it doesn't exist
	 */
  public int jjtmSearchChild(SimpleNode node) {
	  int rslt = -1;
	  for (int i = 0; i < children.length; i++)
		  if (children[i] == node) {
			  rslt = i;
			  break;
		  } 
	  return rslt;
  }
	
	/*
	 * Removes child node.
	 */
/*
	public void jjtmRemoveChild(SimpleNode node) {
		// search node
		boolean found = false;
		int i;
		for (i = 0; i < children.length; i++)
			if (children[i] == node) {
				found = true;
				break;
			}
		// if found remove it
		if (found) {
			Node c[] = new Node[children.length - 1];
			if (i > 0) System.arraycopy(children, 0, c, 0, i);
			if (i < children.length-1) 
				System.arraycopy(children, i+1, c, i, children.length-i-1);
			children = c;
		}
	}
*/
	
	// Other things - unused stuff
	
  /** Symbol table */
//  protected static java.util.Hashtable symtab = new java.util.Hashtable();

  /** Stack for calculations. */
//  protected static Object[] stack = new Object[1024];
//  protected static int top = -1;

/*
  public void interpret()
  {
	 throw new Error(); // It better not come here.
  }
*/
}

